#include <Astri_DAQ.h>
#include <sstream>
#include <iomanip>
#include <unistd.h>
#include <ctime>

#include <Astri_Redis_DAQ_Observer.h>
#include <TCP_Protocol.h>
#include <UDP_Protocol.h>
#include <File_Archiver.h>
#include <Kafka_Provider.h>
#include <Redis_Provider.h>
#include <Astri_Monitor.h>
#include <Astri_MA_Generic.h>
#include <Packet_Structure_Json.h>

#ifndef ASTRI_PACKETS_PATH
#define ASTRI_PACKETS_PATH std::string(std::getenv("HOME")).append("/BIAS/config/AstriPackets")
#endif

using namespace inaf::oasbo;
using namespace inaf::oasbo::DAQ;
using namespace inaf::oasbo::DAQ_observers;

DAQ::AstriDAQ::AstriDAQ() {
	this->receiver = Receivers::TCPProtocolBuilder().getReceiver();
	this->archiver = Archivers::FileArchiverBuilder().getArchiver();
	this->provider = Providers::RedisProviderBuilder().getProvider();
	this->monitor = PacketMonitors::AstriPacketMonitorBuilder().getMonitor();
	std::string ma_packet(
			std::string(ASTRI_PACKETS_PATH).append(
					"/AstriMA/AstriMAGeneric.json"));
	Packets::PacketStructureJson structure(ma_packet);
	this->packet = new Packets::AstriMaGeneric(structure);

	this->currentState = Status::STOP;
	firstIdle = true;
}

std::string DAQ::AstriDAQ::getStateStr(Status state) {
	switch (state) {
	case Status::STOP:
		return std::string("STOP");
	case Status::IDLE:
		return firstIdle ? std::string("FIRST_IDLE") : std::string("IDLE");
	case Status::RUN:
		return std::string("RUN");
	case Status::READY:
		return std::string("READY");
	default:
		return std::string("NONE");
	}
}

void DAQ::AstriDAQ::stop() {
	switchState(Status::STOP);
}

int DAQ::AstriDAQ::getDaqID() {
	return this->daq_id;
}

void DAQ::AstriDAQ::setDaqID(int daq_id) {
	this->daq_id = daq_id;
	this->updateDaqIDinConf();
}

void DAQ::AstriDAQ::incrementDaqID() {
	this->daq_id += 1;
	this->updateDaqIDinConf();
}

void DAQ::AstriDAQ::updateDaqIDinConf() {
	// update run id into the configurations
	std::map<std::string, std::string> newConf;
	newConf[AstriDAQBuilder::daq_id_key] = std::to_string(this->daq_id);
	for (auto conf : this->configurations) {
		conf->insert(newConf, AstriDAQBuilder::config_target);
		conf->pushConfigToSource(AstriDAQBuilder::config_target);
	}
}

void DAQ::AstriDAQ::switchState(Status newState) {
	if (newState == STOP) {
		this->setCurrentState(Status::STOP);
		return;
	}

	std::string dest;
	switch (this->currentState) {
	case Status::READY: {
		switch (newState) {
		case Status::IDLE:
			this->setCurrentState(newState);
			dest = computeNewDestination();
			archiver->setDest(dest);
			archiver->open();
			updateArchiverObservers();
			break;
		default:
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Astri DAQ]\t" << "Switch from "
					<< getStateStr(this->currentState) << " to "
					<< getStateStr(newState) << " not allowed" << std::endl;
			break;
		}
		break;
	}
	case Status::IDLE: {
		firstIdle = false;
		switch (newState) {
		case Status::IDLE:
			archiver->close();
			updateArchiverObservers();
			this->incrementDaqID();
			this->setCurrentState(newState);
			dest = computeNewDestination();
			archiver->setDest(dest);
			archiver->open();
			updateArchiverObservers();
			break;
		case Status::RUN:
			archiver->close();
			updateArchiverObservers();
			this->setCurrentState(newState);
			dest = computeNewDestination();
			archiver->setDest(dest);
			archiver->open();
			updateArchiverObservers();
			break;
		case Status::READY:
			archiver->close();
			updateArchiverObservers();
			this->setCurrentState(Status::READY);
			receiver->connectToClient();
			if (!receiver->isConnectedToClient()) {
				time_t now = time(nullptr);
				std::cerr << "["
						<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
						<< "]\t[Astri DAQ]\t"
						<< "Unable to establish connection" << std::endl;
				switchState(Status::STOP);
				break;
			}
			incrementDaqID();
			monitor->reset();
			this->switchState(IDLE);
			break;
		default:
			break;
		}
		break;
	}
	case Status::RUN: {
		switch (newState) {
		case Status::IDLE:
			archiver->close();
			updateArchiverObservers();
			this->incrementDaqID();
			this->setCurrentState(newState);
			dest = computeNewDestination();
			archiver->setDest(dest);
			archiver->open();
			updateArchiverObservers();
			break;
		case Status::RUN:
			archiver->close();
			updateArchiverObservers();
			this->incrementDaqID();
			dest = computeNewDestination();
			archiver->setDest(dest);
			archiver->open();
			updateArchiverObservers();
			break;
		case Status::READY:
			archiver->close();
			updateArchiverObservers();
			this->setCurrentState(Status::READY);
			receiver->connectToClient();
			if (!receiver->isConnectedToClient()) {
				time_t now = time(nullptr);
				std::cerr << "["
						<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
						<< "]\t[Astri DAQ]\t"
						<< "Unable to establish connection" << std::endl;
				switchState(Status::STOP);
				break;
			}
			incrementDaqID();
			monitor->reset();
			this->switchState(IDLE);
			break;
		default:
			break;
		}
		break;
	}
	case Status::STOP:
		switch (newState) {
		case Status::READY:
			std::for_each(this->observers.begin(), this->observers.end(),
					[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
						ob->start();
					}); // start observers
			if (provider->open() > 0) {
				time_t now = time(nullptr);
				std::cout << "["
						<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
						<< "]\t[Astri DAQ]\t" << "Provider connected"
						<< std::endl;
			}
			this->checkProvThreadFlag = true;
			this->checkProvConn = std::thread(&AstriDAQ::connectProvider, this);
			receiver->connectToClient();
			if (!receiver->isConnectedToClient()) {
				time_t now = time(nullptr);
				std::cerr << "["
						<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
						<< "]\t[Astri DAQ]\t"
						<< "Unable to connect Receiver" << std::endl;
				this->setCurrentState(STOP);
				break;
			} else {
				time_t now = time(nullptr);
				std::cout << "["
						<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
						<< "]\t[Astri DAQ]\t" << "Receiver connected"
						<< std::endl;
			}
			this->setCurrentState(Status::READY);
			this->switchState(IDLE);
			break;
		default:
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Astri DAQ]\t" << "Switch from "
					<< getStateStr(this->currentState) << " to "
					<< getStateStr(newState) << " not allowed" << std::endl;
			break;
		}
		break;
	default:
		break;
	}
}

void DAQ::AstriDAQ::start() {
	incrementDaqID();
	switchState(Status::READY);

	// start receiving
	while (currentState != Status::STOP && currentState != Status::READY) {
		int rec = receiver->receiveFromClient(*this->packet);
		if (rec <= 0) { // error in receiving or closed connection
			if (!receiver->isConnectedToClient()) { // reconnect
				time_t now = time(nullptr);
				std::cout << "["
						<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
						<< "]\t[Astri DAQ]\t" << "Connessione chiusa"
						<< std::endl;
				this->switchState(Status::READY);
			}
			continue;
		}
		if (!this->packet->isRecognizedHeader()) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Astri DAQ]\t" << "Packet header not recognized "
					<< std::endl;
			continue;
		}
		this->deliverPacket();
		Packets::AstriMaGeneric *packet =
				static_cast<Packets::AstriMaGeneric*>(this->packet); // downcast
		if (packet->getType() == 15) { //NOTIF PACKET
			time_t now = time(nullptr);
			std::cout << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Astri DAQ]" << std::endl;
			monitor->printStats();
			std::cout << "-------------------------------" << std::endl;
			unsigned short cmd = ((this->packet->getBinaryPointer()[8]) << 8)
					+ (this->packet->getBinaryPointer()[9]);
			Status newState = cmd == 43690 ? Status::RUN : Status::IDLE;
			this->setNextState(newState);
			now = time(nullptr);
			std::cout << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Astri DAQ]\t" << "Switching from "
					<< getStateStr(this->currentState) << " to -> "
					<< getStateStr(newState) << std::endl;
			this->switchState(getNextState());
		}
	}

	// stop receiving
	if (archiver->is_open()) {
		this->archiver->close();
		updateArchiverObservers();
	}

	this->checkProvThreadFlag = false;
	if (this->checkProvConn.joinable()) {
		this->checkProvConn.join();
	}
	this->provider->close();

	this->receiver->closeConnectionToClient();

	std::for_each(this->observers.begin(), this->observers.end(),
			[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
				ob->stop();
			});

}

int DAQ::AstriDAQ::deliverPacket() {
	archiver->write(*packet);
	provider->isOpen() && provider->write(*packet);
	monitor->monit(*packet);
	return 0;
}

void DAQ::AstriDAQ::updateObservers() {
	std::for_each(this->observers.begin(), this->observers.end(),
			[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
				ob->updateAll();
			});
}

void DAQ::AstriDAQ::updateArchiverObservers() {
	std::for_each(this->observers.begin(), this->observers.end(),
			[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
				ob->updateArchiverStats();
			});
}

int AstriDAQ::connectProvider() {
	while (checkProvThreadFlag) {
		if (!provider->isOpen()) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Astri DAQ]" << "\tProvider not connected"
					<< std::endl;
			if (provider->open() > 0) {
				time_t now = time(nullptr);
				std::cout << "["
						<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
						<< "]\t[Astri DAQ]" << "\tProvider connected"
						<< std::endl;
			}
		}
		std::this_thread::sleep_for(std::chrono::seconds(2));
	}
	return 1;
}

std::string DAQ::AstriDAQ::computeNewDestination() {
	std::string target = AstriDAQBuilder::sched_block_config_target;
	std::map<std::string, std::string> actConf;
	for (auto conf : this->configurations) {
		conf->readConfigFromSource(target);
		for (auto [k, v] : conf->getConfig()) {
			actConf[k] = v;
		}
	}

	std::stringstream ss;
	int decade = this->daq_id / 10;
	ss << std::setfill('0') << std::setw(4) << std::to_string(decade) << "/";
	ss << actConf[target + "_" + AstriDAQBuilder::date_id_key] << "_MA";
	ss << std::setfill('0') << std::setw(2)
			<< actConf[target + "_" + AstriDAQBuilder::tel_id_key] << "_";
	ss << actConf[target + "_" + AstriDAQBuilder::target_id_key] << "_";
	ss << actConf[target + "_" + AstriDAQBuilder::obs_mode_key] << "_";
	ss << std::setfill('0') << std::setw(8)
			<< actConf[target + "_" + AstriDAQBuilder::run_id_key] << "_";
	std::string mode_id;
	if (this->currentState == Status::IDLE)
		mode_id = firstIdle ? "F" : "I";
	else
		mode_id = "R";

	ss << mode_id << "_";
	ss << std::setfill('0') << std::setw(6) << this->daq_id;
	return ss.str();
}

